'''
A base class for RESTful resource implementations to be used as Django views.
Dispatching is handled based on the HTTP method of the request. When handling
a request, the instance-method named exactly like the lowercase version of
the HTTP method is called. The response depends on the control flow of the
specific handler.
If it returns normally, whatever value is returned from the handler is the
response. Otherwise the `Resource` tries to emit a meaningful error message.

- If a `Problem` was raised, it is returned as a response.
- If an ObjectDoesNotExist exception was raised, a `Problem.not_found` instance
  is returned.
- Otherwise the exception is wrapped in a `Problem.internal_error`.

All Problems belonging to the 5XX-class of HTTP errors are logged to stdout.
'''

from django.core.exceptions import ObjectDoesNotExist

from papyru.logger import LogSequence, log_root
from papyru.problem import Problem

# defined in RFC 7231
HTTP_METHODS = set(['get', 'head', 'post', 'put', 'delete', 'connect',
                    'options', 'trace'])


class Resource:
    @log_root()
    def __call__(self, request, *args):
        try:
            try:
                method = getattr(self.__class__, request.method.lower())
            except AttributeError:
                if (request.method.lower() == 'head' and
                        getattr(self.__class__, 'get', False)):
                    method = _simple_head
                else:
                    method = _method_now_allowed(
                        allowed=set(dir(self.__class__)) & HTTP_METHODS)

            def respond():
                return method(self, request, *args)

            should_log_access = getattr(method, '_should_log_access', True)

            if should_log_access:
                request_id = request.headers.get('pap-request-id', None)

                with LogSequence('%s%s %s' % (('[%s] ' % request_id
                                               if request_id is not None
                                               else ''),
                                              request.method,
                                              request.get_full_path())):
                    return respond()
            else:
                return respond()

        except Problem as problem:
            return problem.to_response()

        except ObjectDoesNotExist:
            return Problem.not_found().to_response()

        except NotImplementedError:
            return Problem.not_implemented().to_response()

        except Exception as exc:
            return Problem.internal_error(
                'unexpected error: %s' % exc).to_response()


def no_access_log(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper._should_log_access = False
    return wrapper


def _simple_head(self, request, *args):
    method = getattr(self.__class__, 'get')
    resp = method(self, request, *args)
    resp.content = ''
    return resp


def _method_now_allowed(allowed):
    def wrapped(*args, **kwargs):
        raise Problem.method_not_allowed(allowed)

    return wrapped



'''
When a transport representation is given it can be validated during
construction.

Validation is defined by a validator, given by the validator field of the
serializer. This field can hold a reference to an object that implements a
validate method. When this method is called with a transport representation it
should either return a normalized version of the representation or raise an
error.

Predefined validators are the CerberusValidator and the JSONSchemaValidator.
The CerberusValidator checks objects against a cerberus schema definition. The
JSONSchemaValidator checks against JSON Schemes.
'''

import json
from io import StringIO
from os.path import dirname
from urllib.parse import urlparse

import cerberus
import jsonschema
from lxml import etree
from requests import get

from .problem import Problem


def _validation_error(detail):
    return Problem.unsupported_media_type(detail=detail)


class Validator:
    def validate(self, representation):
        raise NotImplementedError()


class JSONSchemaValidator(Validator):
    def __init__(self, spec_file_name, format_checker=None):
        self.format_checker = format_checker
        with open(spec_file_name, 'r') as f:
            self.schema = json.load(f)

    def validate(self, representation):
        try:
            jsonschema.validate(instance=representation,
                                schema=self.schema,
                                format_checker=self.format_checker)
            return representation
        except jsonschema.exceptions.ValidationError as exc:
            raise _validation_error('%s' % exc)


class CerberusValidator(Validator):
    def __init__(self, schema_description):
        self.validator = cerberus.Validator(schema_description['schema'])

        if 'allow_unknown' in schema_description:
            self.validator.allow_unknown = schema_description['allow_unknown']

    def validate(self, representation):
        try:
            if not self.validator.validate(representation):
                raise _validation_error('%s' % self.validator.errors)
            else:
                return self.validator.normalized(representation)
        except cerberus.validator.DocumentError as exc:
            raise _validation_error('%s' % exc)


class XSDValidator(Validator):
    def __init__(self, xsd_path):
        def _resolve_url(path):
            response = get(path)
            response.raise_for_status()

            if not any(map(lambda content_type: (response
                                                 .headers['Content-Type']
                                                 .startswith(content_type)),
                           {'text/xsd', 'text/xml', 'application/xml'})):
                raise ValueError('unexpected content type: `%s`'
                                 % response.headers['Content-Type'])

            return response.text

        xsd_base_directory = '%s/' % dirname(xsd_path)

        self.parser = etree.XMLParser()

        class UrlResolver(etree.Resolver):
            def resolve(self, referenced_location, id, context):
                if (urlparse(referenced_location).scheme not in
                        {'http', 'https'}):
                    return None

                return self.resolve_string(
                    _resolve_url(referenced_location),
                    context,
                    base_url=xsd_base_directory)

        self.parser.resolvers.add(UrlResolver())

        doc = etree.parse(xsd_path, parser=self.parser,
                          base_url=xsd_base_directory)
        self.xml_schema = etree.XMLSchema(doc)

    def validate(self, representation):
        try:
            self.xml_schema.assertValid(
                etree.parse(
                    StringIO(representation)))
            return representation
        except etree.DocumentInvalid as exc:
            raise _validation_error('%s' % exc)



import json
import os
import time
from functools import reduce
from random import randint
from time import perf_counter

from .context import JSONContext
from .resource import Resource


class ProfilerMeasurement:
    def __init__(self, profiler, item):
        self._profiler = profiler
        self._item = item

    def __enter__(self):
        self._start = perf_counter()
        self._sub_profiler = Profiler(self._profiler.sub_id(self._item))
        return self._sub_profiler

    def __exit__(self, *args):
        self._profiler._add_measurement(self._item,
                                        perf_counter() - self._start)
        self._profiler.merge(self._sub_profiler._measurements)


def DummyProfiler():
    return Profiler('dummy')


class Profiler:
    def __init__(self, id):
        self._id = id
        self._measurements = []

    def measure(self, item):
        return ProfilerMeasurement(self, item)

    def _add_measurement(self, item, time):
        def milliseconds(seconds):
            return int(seconds * 1000)

        self._measurements.append(('%s/%s' % (self._id, item),
                                   milliseconds(time)))

    def sub_id(self, sub_id):
        return '%s/%s' % (self._id, sub_id)

    def pack(self, value):
        return (self._measurements, value)

    def unpack(self, packed):
        self.merge(packed[0])
        return packed[1]

    def merge(self, measurements):
        self._measurements = self._measurements + measurements

    def save_report(self, reports_dir):
        filename = '%s/%s_%s_%s' % (reports_dir,
                                    self._id,
                                    time.strftime(
                                        '%Y-%m-%d-%H-%M-%S',
                                        time.gmtime()),
                                    randint(0, 99999))

        try:
            os.makedirs(reports_dir)
        except FileExistsError:
            pass

        with open(filename, 'w') as f:
            json.dump(self._format_report(), f)

    def _format_report(self):
        def group_measurements(acc, el):
            item, dt = el

            if item in acc:
                acc[item] = acc[item] + dt
            else:
                acc[item] = dt

            return acc

        return reduce(group_measurements, self._measurements, {})


class ReportsResource(Resource):
    def __init__(self, reports_dir):
        self._reports_dir = reports_dir


class ReportList(ReportsResource):
    def get(self, request):
        return JSONContext().response(os.listdir(self._reports_dir))


def _get_report_filenames(reports_dir, prefix):
    return list(filter(
        lambda filename: filename.startswith(prefix),
        os.listdir(reports_dir)))


class ReportCSVDetails(ReportsResource):
    def get(self, request, prefix):
        def format_csv(obj):
            return '\n'.join(list(
                map(lambda it: '"%s", %s' % it,
                    sorted(obj.items(),
                           key=lambda it: it[1],
                           reverse=True))))

        def read_report(acc, filename):
            with open('%s/%s' % (self._reports_dir, filename)) as f:
                acc[filename] = format_csv(json.load(f))

            return acc

        reports = reduce(read_report, _get_report_filenames(self._reports_dir,
                                                            prefix), {})
        return JSONContext().response(reports)


class ReportPlotDetails(ReportsResource):
    def get(self, request, prefix):
        def shortname(txt):
            CUTOFF = 50

            return (
                '…%s' % txt[-(min(len(txt), CUTOFF)):]
                if len(txt) > CUTOFF
                else txt)

        def is_significant(item):
            TIME_THRESHOLD = 1  # ms
            return item[1] > TIME_THRESHOLD

        def format_gnuplot(obj):
            return '\n'.join(list(
                map(lambda it: '"%s" %s' % (shortname(it[0]), it[1]),
                    filter(is_significant,
                           sorted(obj.items(),
                                  key=lambda it: it[0])))))

        def read_report(acc, filename):
            with open('%s/%s' % (self._reports_dir, filename)) as f:
                acc[filename] = format_gnuplot(json.load(f))

            return acc

        reports = reduce(read_report, _get_report_filenames(self._reports_dir,
                                                            prefix), {})
        return JSONContext().response(reports)


def should_profile_request(request):
    return request.META.get('HTTP_PROFILE', 'false') == 'true'



import inspect


class IncompatiblePythonImplementation(Exception):
    pass


class MissingStackContext(Exception):
    pass


_CONTEXT_LOOKUP = {}


def put_stackcontext(key, value):
    def wrapper(fn):
        def impl(*args, **kwargs):
            frame = inspect.currentframe()

            if frame is None:
                raise IncompatiblePythonImplementation()

            try:
                context = _CONTEXT_LOOKUP.get(frame, {})

                context[key] = (value()
                                if inspect.isfunction(value)
                                else value)

                _CONTEXT_LOOKUP[frame] = context

                return fn(*args, **kwargs)
            finally:
                del context[key]

                if len(context) == 0:
                    del _CONTEXT_LOOKUP[frame]

                del frame
        return impl
    return wrapper


def get_stackcontext(key, fallback=None):
    ptr = inspect.currentframe()

    while ptr is not None:
        if ptr in _CONTEXT_LOOKUP:
            context = _CONTEXT_LOOKUP[ptr]
            if key in context:
                return context[key]

        ptr = ptr.f_back

    if fallback is not None:
        return fallback

    raise MissingStackContext()



#!/usr/bin/env python3

import json
import sys
from functools import reduce
from os.path import join

import yaml


def _load_schemas_from_file(schema_file):
    with open(schema_file) as f:
        data = yaml.load(f.read(), Loader=yaml.SafeLoader)
    return data['components']['schemas']


def _load_ref(value, schemas, schemas_dir):
    custom_schema_file = value.split('#')[0]
    schema_to_load = value.split('/')[-1]
    if custom_schema_file == '':
        sub_schemas = schemas
    else:
        sub_schemas = _load_schemas_from_file(join(schemas_dir,
                                                   custom_schema_file))
    return _load_references(sub_schemas[schema_to_load],
                            sub_schemas,
                            schemas_dir)


def _load_list(schema, schemas, schemas_dir):
    for index, value in enumerate(schema):
        schema[index] = _load_references(value, schemas, schemas_dir)

    return schema


def _load_dict(schema, schemas, schemas_dir):
    for key in list(schema):
        if '$ref' in key:
            schema.update(_load_ref(schema[key], schemas, schemas_dir))
            del schema[key]
        elif key == 'enum' and 'items' in schema.keys():
            schema['items']['enum'] = schema['enum']
            del schema['enum']
        elif key == 'x-extensible-enum' and 'items' in schema.keys():
            schema['items']['enum'] = schema['x-extensible-enum']
            del schema['x-extensible-enum']
        elif key == 'x-extensible-enum':
            schema['enum'] = schema['x-extensible-enum']
            del schema['x-extensible-enum']
        else:
            schema[key] = _load_references(schema[key], schemas, schemas_dir)
    return schema


def _load_references(schema, schemas, schemas_dir):
    if type(schema) is list:
        schema = _load_list(schema, schemas, schemas_dir)
    elif type(schema) is dict:
        schema = _load_dict(schema, schemas, schemas_dir)
    return schema


def resolve_inheritance(schemas):
    if 'allOf' in schemas:
        def merge_dict_items(acc, el):
            key, value = el
            acc[key] = value
            return acc

        def merge_values(lhs, rhs):
            if isinstance(lhs, list):
                return list(set(lhs) | set(rhs))
            elif isinstance(lhs, dict):
                return reduce(merge_dict_items, lhs.items(), rhs)
            else:
                raise NotImplementedError()

        def combine_schemas(acc, el):
            for key, value in el.items():
                if key in acc:
                    acc[key] = merge_values(acc[key], value)
                else:
                    acc[key] = value
            return acc

        return reduce(combine_schemas, schemas['allOf'], {})
    else:
        return schemas


def main(argv):
    if len(argv) < 3:
        print('usage: %s SWAGGER_FILE OUTPUT_DIR' % argv[0])
        sys.exit(1)
    else:
        schemas_dir = argv[1].rsplit('/', 1)[0]
        schemas = _load_schemas_from_file(argv[1])
        for title, schema in schemas.items():
            json_schema = {
                **resolve_inheritance(
                    _load_references(schema, schemas, schemas_dir)),
                'additionalProperties': False
            }

            with open('%s/%s.json' % (argv[2], title),
                      'w') as of:
                of.write(json.dumps(json_schema, indent=2))


if __name__ == '__main__':
    main(sys.argv)



import json
import signal
import traceback
from datetime import timedelta
from functools import partial, wraps
from http import HTTPStatus
from multiprocessing import Pool
from os import environ, getpid
from sys import exit
from threading import Condition
from time import sleep
from typing import Callable, List, Optional, Union
from urllib.parse import urljoin

import requests
from django.core.management.base import BaseCommand

from papyru.utils import limited_runtime, log
from papyru.varspool.config import VarspoolJobProcessorConfig
from papyru.varspool.types import (Job, JobResult, JobStatus, JobStatusEntry,
                                   JobStatusHistory)


class JobCounter:
    def __init__(self):
        self._counter = 0
        self._active_jobs = dict()
        self._cv = Condition()

    @property
    def active_jobs_count(self):
        return self._counter

    @property
    def active_jobs_ids(self):
        return list(self._active_jobs.keys())

    @property
    def active_jobs(self):
        return list(self._active_jobs.values())

    def inc(self, job: Job):
        with self._cv:
            self._counter += 1
            self._active_jobs[job.id] = job
            self._cv.notify()

    def dec(self, job: Union[Job, int]):
        with self._cv:
            self._counter -= 1

            if self._counter < 0:
                raise AssertionError("job counter must not be negative")

            key = (job
                   if isinstance(job, int)
                   else job.id)
            del self._active_jobs[key]

            self._cv.notify()

    def wait_for_change(self, timeout: float):
        with self._cv:
            self._cv.wait(timeout=timeout)


def loop_until(time_range: timedelta):
    '''
    Decorates a function which performs a single loop step.

    Calls the function as long as the according process has neither
    received SIGTERM nor the given time range has been exceeded.
    '''

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            state = {'received_sigterm': False}

            def handle_sigterm(*args, **kwargs):
                log('caught SIGTERM. will shutdown gracefully.')
                state['received_sigterm'] = True

            signal.signal(signal.SIGTERM, handle_sigterm)

            with limited_runtime(time_range) as has_runtime_left:
                while has_runtime_left() and not state['received_sigterm']:
                    func(*args, **kwargs)
        return wrapper
    return decorator


g_job_counter = JobCounter()


def _err_callback(job: Job, ex: Exception):
    if isinstance(ex, CouldNotProcessJobException):
        log(ex.traceback, level='error')
        log('[ALERT] could not process job %d: %s' % (ex.job_id, ex.message),
            level='error')
    elif isinstance(ex, StatusUpdateConflict):
        log('[ALERT] could not finish job %d due to status conflict'
            % ex.job_id, level='error')
    else:
        log('[ALERT] internal error processing job due to %s: %s'
            % (type(ex).__name__, str(ex)),
            level='error')

    g_job_counter.dec(job)


def _succ_callback(job: Job):
    log('finished job %d with status %s.'
        % (job.id, job.status_history.items[-1].status))
    g_job_counter.dec(job)


def _get_num_of_jobs_to_fetch(max_parallel_job_count: int):
    return max_parallel_job_count - g_job_counter.active_jobs_count


def _fetch_jobs(queue_url: str, limit: int) -> List[Job]:
    endpoint = ('jobs?current_status=%s&limit=%d'
                % (str(JobStatus.OPEN), limit))

    resp = requests.get(urljoin('%s/' % queue_url, endpoint))
    resp.raise_for_status()

    return list(map(Job.from_dict, resp.json()['jobs']))


def _post_job_status(job: Job, new_status: JobStatusEntry) -> Job:
    previous_status = job.status_history.items[-1]

    resp = requests.post(
        job.status_history.location,
        data=json.dumps({
          'previous_status': {
              'checksum': str(previous_status.checksum),
          },
          **new_status.to_dict()}))

    if resp.status_code == HTTPStatus.CONFLICT:
        raise StatusUpdateConflict(job.id)
    elif resp.status_code != HTTPStatus.OK:
        raise CouldNotPersistJobStatusException(resp.status_code, resp.text)

    return Job(id=job.id,
               status_history=JobStatusHistory(
                   location=job.status_history.location,
                   items=list(map(JobStatusEntry.from_dict,
                                  resp.json()['items']))))


def handle_job(job: Job,
               processing_fn: Callable) -> Job:
    '''
    A wrapper for a function processing a job.

    This wrapper function does all the bureaucratic stuff which needs to be
    done around the hard work. It notifies varspool that the job has been
    started and sends the result status after the job has been finished.
    '''

    log('processing job %d.' % job.id)

    try:
        # tell varspool the job has been started and provide some debug
        # information
        job = _post_job_status(
            job,
            JobStatusEntry(
                status=JobStatus.IN_PROGRESS,
                data={
                    'pod': (
                        environ.get(
                            'HOSTNAME',
                            'could not determine pod.')),
                    'worker-pid': getpid(),
                }))

    except StatusUpdateConflict:
        log('job %d is already in progress. skipping.' % job.id)
        return job
    except Exception:
        raise CouldNotProcessJobException(
            job.id,
            'failed telling varspool the job is in progress',
            traceback.format_exc())

    try:
        result = processing_fn(job)

    except Exception as ex:
        formatted_traceback = traceback.format_exc()
        formatted_exc = ('uncaught error (%s): %s'
                         % (type(ex).__name__, str(ex)))

        log(formatted_traceback, level='error')
        log(formatted_exc, level='error')

        result = JobResult(status=JobStatus.FAILED,
                           data={
                               'reason': formatted_exc,
                               'traceback': formatted_traceback,
                           })

    log('processed job %d. updating varspool.' % job.id)

    try:
        # update varspool with the fallback failure result or the processing
        # result
        job = _post_job_status(job, result.to_status_entry())

    except StatusUpdateConflict:
        raise
    except Exception:
        raise CouldNotProcessJobException(
            job.id,
            'failed to persist job status. result was: %s' % result,
            traceback.format_exc())

    return job


def wait_for_free_slot(timeout: float):
    g_job_counter.wait_for_change(timeout)


def enter_job_loop(config: VarspoolJobProcessorConfig):
    '''
    Main job loop launching a process pool and asynchronously calling
    job processing functions.
    '''

    try:
        log('checking config...')

        if not isinstance(config, VarspoolJobProcessorConfig):
            raise RuntimeError(
                'could not start job loop due to invalid config.')

        log('- queue url: %s' % config.queue_url)

        with Pool(processes=config.max_parallel_job_count) as pool:
            @loop_until(
                time_range=timedelta(minutes=config.max_runtime_in_minutes))
            def wrapper():
                log('fetching jobs...')

                if (g_job_counter.active_jobs_count
                        >= config.max_parallel_job_count):
                    log('too many active jobs. waiting for a free slot.')
                    wait_for_free_slot(config.loop_cooldown_in_seconds)
                    return

                try:
                    jobs = _fetch_jobs(config.queue_url,
                                       _get_num_of_jobs_to_fetch(
                                           config.max_parallel_job_count))
                except Exception as exc:
                    raise FetchJobException('%s: %s' % (type(exc).__name__,
                                                        str(exc)))

                if len(jobs) == 0:
                    log('no jobs to process. waiting...')
                    sleep(config.loop_cooldown_in_seconds)
                    return

                for job in jobs:
                    log('dispatching job %d.' % job.id)
                    g_job_counter.inc(job)

                    pool.apply_async(
                        func=handle_job,
                        args=(job,
                              config.job_handler),
                        callback=_succ_callback,
                        error_callback=partial(_err_callback, job))
                    log('dispatched job %d.' % job.id)
            wrapper()

            # time is over or SIGTERM has been fired to the job processor.
            # abort our work.
            if g_job_counter.active_jobs_count > 0:
                log('aborting %d unfinished job(s): %s'
                    % (g_job_counter.active_jobs_count,
                       ', '.join(map(str, g_job_counter.active_jobs_ids))))

            for job in g_job_counter.active_jobs:
                try:
                    job = _post_job_status(
                        job,
                        JobStatusEntry(
                            status=config.abort_status,
                            data={
                                'reason': 'graceful shutdown',
                            }))
                    log('aborted job %d' % job.id)
                except Exception:
                    log('could not propagate job termination to varspool '
                        '(job id: %d)' % job.id)

    except Exception as ex:
        log(traceback.format_exc(), level='error')
        log('[ALERT] critical error occurred: %s (%s). '
            'awaiting end of failure backoff...'
            % (type(ex).__name__, str(ex)),
            level='error')

        sleep(config.failure_backoff_in_seconds
              if hasattr(config, 'failure_backoff_in_seconds')
              else (VarspoolJobProcessorConfig
                    .DEFAULT_FAILURE_BACKOFF_IN_SECONDS))
        exit(1)


class VarspoolJobProcessorBaseCommand(BaseCommand):
    config: VarspoolJobProcessorConfig = None

    def handle(self, *args, **kwargs):
        log('varspool job processor started.')

        log('entering job loop.')
        enter_job_loop(self.config)
        log('goodbye.')


class FetchJobException(Exception):
    pass


class StatusUpdateConflict(Exception):
    def __init__(self, job_id: int):
        self.job_id = job_id


class CouldNotPersistJobStatusException(Exception):
    def __init__(self, status_code: int, text: str):
        self.status_code = status_code
        self.text = text


class CouldNotProcessJobException(Exception):
    def __init__(self,
                 job_id: int, message: str, traceback: Optional[str] = None):
        self.job_id = job_id
        self.message = message
        self.traceback = traceback



from collections import namedtuple


class Message:
    SUCCESS = 1
    INFO = 2
    WARNING = 3
    FAILURE = 4
    CRITICAL = 5

    def __init__(self, severity, message):
        self.severity = severity
        self.message = message

    def __eq__(self, other):
        if isinstance(other, Message):
            return (self.severity == other.severity and
                    self.message == other.message)
        else:
            return NotImplemented


class Sequence:
    def __init__(self, description, resolution=None, items=None):
        self.description = description
        self.resolution = resolution
        self.items = (items
                      if items is not None
                      else [])

    def append(self, item):
        self._items.append(item)

    def __eq__(self, other):
        if isinstance(other, Sequence):
            return (self.description == other.description and
                    self.resolution == other.resolution and
                    len(self.items) == len(other.items) and
                    all(map(lambda it: it[0] == it[1],
                            zip(self.items, other.items))))
        else:
            return NotImplemented


class Trace:
    Summary = namedtuple('Summary', ['description', 'exception'])
    Frame = namedtuple('Frame', ['filename', 'linenumber', 'line'])

    def __init__(self, summary, frames):
        self.summary = summary
        self.frames = frames

    def __eq__(self, other):
        if isinstance(other, Trace):
            return (self.summary == other.summary and
                    self.frames == other.frames)
        else:
            return NotImplemented



import xml.etree.ElementTree as ET
from enum import Enum
from functools import reduce
from hashlib import sha1
from io import StringIO


class XMLHelper:
    def __init__(self, encoded_xml):
        self._namespaces = (
            _avoid_namespace_collisions(
                map(lambda el: el[1],
                    ET.iterparse(StringIO(encoded_xml),
                                 events=['start-ns']))))

        self._document = ET.parse(StringIO(encoded_xml))

        self._register_current_namespaces()

    @property
    def root(self):
        return self._document.getroot()

    def namespace(self, shortname):
        return self._namespaces.get(shortname, None)

    def deep_copy(self):
        return XMLHelper(self.serialize(self.root).decode('utf-8'))

    @property
    def default_namespace(self):
        return self._namespaces.get('', None)

    def attribute_fullname(self, name, namespace=None):
        return ('{%s}%s' % (self.namespace(namespace), name)
                if namespace is not None
                else name)

    def get_attribute(self, node, name, namespace=None):
        try:
            return node.attrib[self.attribute_fullname(name, namespace)]
        except KeyError:
            return None

    def set_attribute(self, node, name, value, namespace=None):
        node.attrib[self.attribute_fullname(name, namespace)] = value

    def del_attribute(self, node, name, namespace=None):
        del node.attrib[self.attribute_fullname(name, namespace)]

    def has_attribute(self, node, name, namespace=None):
        return self.attribute_fullname(name, namespace) in node.attrib

    def serialize(self, node):
        self._register_current_namespaces()
        return ET.tostring(node)

    def _register_current_namespaces(self):
        # ElementTree registers namespaces global. We don't want that!
        # Different SVGs could have different namespaces for the same prefixes!
        for namespace in self._namespaces.items():
            ET.register_namespace(*namespace)


def _collision_free_name(namespace_value):
    return 'namespaceCollision%s' % sha1(
        namespace_value.encode('utf-8')).hexdigest()


def _avoid_namespace_collisions(kv_iter):
    Action = Enum('Action', 'Add Replace')

    def get_conflict_free_key(acc, key, value):
        XML_NAMESPACE_NAME = 'http://www.w3.org/XML/1998/namespace'
        XML_NAMESPACE_PREFIX = 'xml'

        existing_key, _ = next(
            filter(lambda el: el[1] == value, acc.items()),
            (None, None))

        if value == XML_NAMESPACE_NAME:
            return (Action.Add, XML_NAMESPACE_PREFIX)
        elif key.startswith('xml'):
            return (Action.Add, _collision_free_name(value))
        elif existing_key is not None:
            if key == '' and (key not in acc):
                return (Action.Replace, (existing_key, ''))
            else:
                return (Action.Add, existing_key)
        elif key in acc:
            return (Action.Add, _collision_free_name(value))
        else:
            return (Action.Add, key)

    def add_to_dict(state, kv):
        key, value = kv
        action, new_key = get_conflict_free_key(state, key, value)

        if action == Action.Add:
            state[new_key] = value
        elif action == Action.Replace:
            old_key, replacement = new_key
            del state[old_key]
            state[replacement] = value

        return state

    return reduce(add_to_dict, kv_iter, {})



'''
Provides an implementation of a job processor using the puzzle & play
varspool API which dispatches jobs from a specific queue based on a JSON
configuration.

The job processor can both be invoked by defining a Django management command
or directly calling the job loop. It requires a configuration
(:py:class:`papyru.varspool.config.VarspoolJobProcessorConfig`) and a
processing function. The processing function takes a
:py:class:`papyru.varspool.types.Job` and should return a
:py:class:`papyru.varspool.types.JobResult`.

Example:

  - via Management Command
    (:py:class:`papyru.varspool.command.VarspoolJobProcessorBaseCommand`)

    .. code-block:: python

      class Command(VarspoolJobProcessorBaseCommand):
          config = VarspoolJobProcessorConfig(config={...})

  - via directly calling the job loop
    (:py:func:`papyru.varspool.command.enter_job_loop`)

    .. code-block:: python

      enter_job_loop(
          VarspoolJobProcessorConfig(config={...}))
'''

from .command import (VarspoolJobProcessorBaseCommand, enter_job_loop,
                      handle_job, loop_until)
from .config import VarspoolJobProcessorConfig
from .types import Job, JobResult, JobStatus, JobStatusEntry, JobStatusHistory

__all__ = [
    # classes
    'VarspoolJobProcessorConfig',
    'VarspoolJobProcessorBaseCommand',

    # types
    'Job', 'JobResult', 'JobStatus', 'JobStatusEntry', 'JobStatusHistory',

    # functions
    'enter_job_loop', 'handle_job', 'loop_until',
]



import re
from base64 import b64decode
from http import HTTPStatus
from inspect import signature

from .context import PlainContext

ANY_USER = 0x0


def require_basic_auth(user_map, realm='auth'):
    def wrapper(view):
        if user_map is None:
            return view

        get_request = _make_request_argument_binder(view)

        def impl(*args, **kwargs):
            request = get_request(*args, **kwargs)

            request.auth_user = _authenticate(
                user_map, parse_authorization_headers(request))

            if request.auth_user is None:
                auth_request = 'Basic realm="%s", charset="UTF-8"' % realm

                return (PlainContext(content_type=None)
                        .response(body='',
                                  status=HTTPStatus.UNAUTHORIZED,
                                  headers={'WWW-Authenticate': auth_request}))

            return view(*args, **kwargs)
        return impl
    return wrapper


def _make_request_argument_binder(func):
    sig = signature(func)

    def binder(*args, **kwargs):
        bound = sig.bind(*args, **kwargs)

        if 'request' in bound.arguments:
            return bound.arguments['request']
        else:
            return args[0]
    return binder


def _authenticate(user_map, credentials):
    if credentials is None:
        return None

    username, password = credentials

    if user_map == ANY_USER:
        return username

    if username not in user_map:
        return None

    if user_map[username] != password:
        return None

    return username


def parse_authorization_headers(request):
    if 'Authorization' not in request.headers:
        return None

    m = re.match(r'^Basic (.+)$', request.headers['Authorization'])
    if not m:
        return None

    try:
        encoded_credentials = b64decode(m.group(1)).decode('utf-8')
    except Exception:
        return None

    credentials = encoded_credentials.split(':')

    if len(credentials) != 2:
        return None

    return credentials



from django.middleware.common import CommonMiddleware as DjangoCommonMiddleware


class CommonMiddleware(DjangoCommonMiddleware):
    def process_response(self, request, response):
        super().process_response(request, response)

        if (request.method == 'HEAD' and
                (response.content is None or response.content == b'')):
            del response['Content-Length']

        return response



'''
Problems are objects representing some failure that causes the processing of
the current request to be canceled. They are intended to be reported to the
user given an HTTP error code, title and description. When used with Resources,
they can be thrown as exceptions.

You can construct a Problem object for every HTTP error code defined in
`http.HTTPStatus` by calling `Problem.error_status_code_in_lowercase(...)`.

Examples:

- `HTTPStatus.BAD_REQUEST` (400):

  .. code-block:: python

    raise Problem.bad_request(detail="I don't like it.")

- `HTTPStatus.INTERNAL_SERVER_ERROR` (500):

  - **DEPRECATED**:

     .. code-block:: python

       raise Problem.internal_error(detail="Sorry, I still don't like it.")

  - recommended:

     .. code-block:: python

       raise Problem.internal_server_error(
           detail="Sorry, I still don't like it.")

- `HTTPStatus.NOT_IMPLEMENTED` (501):

  .. code-block:: python

    raise Problem.not_implemented()
'''

from http import HTTPStatus
from inspect import getmembers, isfunction
from types import new_class
from typing import Dict
from warnings import warn

from django.http import JsonResponse


def make_problem_class():
    class ProblemImpl(Exception):
        def __init__(self,
                     title: str,
                     status: HTTPStatus,
                     detail: str,
                     should_log: bool = False,
                     headers: Dict[str, str] = {}
                     ):
            self.title = title
            self.status = status
            self.detail = detail
            self.should_log = should_log
            self.headers = headers

        def to_response(self):
            resp = JsonResponse({'title': self.title,
                                 'status': self.status,
                                 'detail': self.detail})
            for key, value in self.headers.items():
                resp[key] = value
            resp.status_code = self.status
            return resp

        # -------------------------------------------------------------------
        # Methods with special features.
        # -------------------------------------------------------------------

        def method_not_allowed(allowed=None):
            headers = {}
            if allowed is not None:
                headers['Allow'] = ', '.join(allowed)
            return Problem('Method not Allowed',
                           HTTPStatus.METHOD_NOT_ALLOWED,
                           ('The requested method is not available on '
                            'this resource'),
                           should_log=False,
                           headers=headers)

        # -------------------------------------------------------------------
        # Methods which stay for backwards compatiblity.
        # -------------------------------------------------------------------

        def not_acceptable():
            return Problem(
                'Not Acceptable',
                HTTPStatus.NOT_ACCEPTABLE,
                ('The requested resource is not capable of generating '
                 'content acceptable according to the Accept headers'))

        def internal_error(detail='internal error'):
            warn('Problem.internal_error() is deprecated. '
                 'Use Problem.internal_server_error() instead.',
                 DeprecationWarning)

            return Problem('Internal Error',
                           HTTPStatus.INTERNAL_SERVER_ERROR,
                           detail,
                           should_log=True)

    overrides = map(lambda it: it[0],
                    getmembers(ProblemImpl, predicate=isfunction))
    error_statuses = dict(map(
        lambda st: (st.name.lower(), st),
        filter(lambda st: (st.value >= 400 and st.value <= 599), HTTPStatus)))

    for method_name in set(error_statuses.keys()) - set(overrides):
        status = error_statuses[method_name]

        def make_problem_function(status, method_name):
            def impl(detail=status.description):
                additionals = {}

                if status.value >= 500 and status.value <= 599:
                    additionals['should_log'] = True

                return Problem(status.phrase, status, detail, **additionals)
            return impl

        setattr(ProblemImpl,
                method_name,
                make_problem_function(status, method_name))

    return new_class('Problem', (ProblemImpl,))


Problem = make_problem_class()



import logging as loglevel
import traceback
from collections import namedtuple
from logging import Handler
from os import path
from threading import current_thread

from papyru.logger_stdout import make_stdout_sink
from papyru.logger_types import Message, Sequence, Trace
from papyru.stack_context import get_stackcontext, put_stackcontext

_STACK_MARKER = '__papyru_log'


def log_root(sink=None):
    def wrapper(fn):
        @put_stackcontext(_STACK_MARKER,
                          lambda: LogRoot(sink or _make_default_sink()))
        def impl(*args, **kwargs):
            return fn(*args, **kwargs)
        return impl
    return wrapper


class PapyruHandler(Handler):
    def emit(self, record):
        _get_root().put(Message(
            severity=PapyruHandler._assess_severity(record.levelno),
            message=self.format(record)))

    def _assess_severity(level):
        if level == loglevel.CRITICAL:
            return Message.CRITICAL
        elif level == loglevel.ERROR:
            return Message.FAILURE
        elif level == loglevel.WARNING:
            return Message.WARNING
        else:
            return Message.INFO


class SequenceGuard:
    def __init__(self, description):
        self.description = description

    def __enter__(self):
        _get_root().save(Sequence(description=self.description,
                                  resolution=Message.SUCCESS))
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        root = _get_root()

        if exc_type is not None:
            root.head.resolution = Message.FAILURE
            _get_root().put(_to_trace(exc_value))

        root.restore()


class LogRoot:
    Context = namedtuple('Context', ['items'])

    def __init__(self, sink):
        self._stack = [LogRoot.Context([])]
        self._sink = sink

    def save(self, context):
        self._stack.append(context)

    def restore(self):
        top = self._stack.pop()
        return self.put(top)

    def put(self, item):
        self.head.items.append(item)
        return self._commit()

    @property
    def head(self):
        return self._stack[-1]

    def _commit(self):
        if len(self._stack) == 1:
            for item in self.head.items:
                self._sink.commit(item)

            self._stack = [LogRoot.Context([])]


class LogSink:
    def commit(self, item):
        raise NotImplementedError()


def _make_default_sink():
    return make_stdout_sink()


def _make_default_root():
    return LogRoot(_make_default_sink())


_thread_roots = {}


def _get_root():
    thread_id = current_thread().ident

    if thread_id not in _thread_roots:
        _thread_roots[thread_id] = _make_default_root()

    return get_stackcontext(_STACK_MARKER, _thread_roots[thread_id])


def _to_trace(exception):
    trace = traceback.TracebackException.from_exception(exception)

    return Trace(
        Trace.Summary(type(exception).__name__, str(exception)),
        list(map(lambda frame: Trace.Frame(path.basename(frame.filename),
                                           frame.lineno,
                                           frame.line),
                 trace.stack)))


# === LEGACY API -- Use the PapyruHandler instead! ============================


def log_item(message):
    return _get_root().put(Message(Message.INFO, message))


def log_info(message):
    return _get_root().put(Message(Message.INFO, message))


def log_ok(message):
    return _get_root().put(Message(Message.SUCCESS, message))


def log_fail(message):
    return _get_root().put(Message(Message.FAILURE, message))


def log_warn(message):
    return _get_root().put(Message(Message.WARNING, message))


def log_trace(exception):
    return _get_root().put(_to_trace(exception))


class LogSequence(SequenceGuard):
    '''
    The LogSequence can be used to create an easy human readable logging trace.

    **Example Call:**

    .. code-block:: python

      with LogSequence('foobar') as log:
          log.info('foobar started')
          log.warn('something is not perfect')
          with LogSequence('bar') as lo:
              lo.ok('child logging')
          with LogSequence('foo') as lo:
              lo.fail('foo failed')
              raise Exception('Something went wrong..')

    **Example Output:**

    ::

      ⚫ foobar
         - foobar started
         ⚠ something is not perfect
         ⚫ bar
            ✓ child logging
            ☀ succeeded
         ⚫ foo
            ✗ foo failed
            ☇ failed
               ↪ foobar.py l.11 | raise Exception('Something went wrong..')
               ↪ Exception: Something went wrong..
         ☇ failed
            ↪ foobar.py l.11 | raise Exception('Something went wrong..')
            ↪ Exception: Something went wrong..
    '''

    def ok(self, message):
        return log_ok(message)

    def fail(self, message):
        return log_fail(message)

    def warn(self, message):
        return log_warn(message)

    def info(self, message):
        return log_info(message)



'''
A context parses input from the client and creates HTTP responses given a
response object, status and headers. The input and response object are encoded
and decoded according to the context.

Example:

    .. code-block:: python

        class ExampleResource(Resource):
            def post(self, request):
                ctx = JSONContext(request)

                return ctx.response(
                    body={'echo': ctx.data['user_message'],
                          status=HTTPStatus.CREATED,
                          headers={'some-header': 'example'}})
'''

import json
from http import HTTPStatus

from django.http import HttpResponse

from .problem import Problem
from .xml_helper import XMLHelper


class RequestContext:
    def __init__(self, request=None):
        if request is None:
            self.request = None
            self.data = None

        else:
            self.request = request

            try:
                self.data = self.decode(request.body)
            except Problem as p:
                raise p
            except Exception as ex:
                raise Problem.unsupported_media_type(
                    detail='%s: %s' % (type(ex).__name__, str(ex)))

    def response(self, body=None, status=HTTPStatus.OK, headers={}):
        content = self.encode(body) if body is not None else ''
        resp = HttpResponse(content=content,
                            content_type=self.content_type,
                            status=status)

        for key, value in headers.items():
            resp[key] = value

        return resp


class PlainContext(RequestContext):
    def __init__(self, content_type):
        self.content_type = content_type

    def encode(self, data):
        return data

    def decode(self, data):
        return data


class JSONContext(RequestContext):
    def decode(self, json_data):
        obj = json.loads(json_data)

        if not isinstance(obj, dict):
            raise Problem.bad_request(
                detail='toplevel JSON value must be an object')

        return obj

    def encode(self, data):
        return json.dumps(data) if data is not None else ''

    @property
    def content_type(self):
        return 'application/json'


class XMLContext(RequestContext):
    def decode(self, encoded_xml_data):
        try:
            return XMLHelper(encoded_xml_data.decode('utf-8'))
        except Exception:
            raise Problem.bad_request()

    def encode(self, xml):
        return xml.serialize(xml.root)

    @property
    def content_type(self):
        return 'application/xml'



'''
Serializers are used to translate objects from transport to internal
representation and back. Both representations can be chosen freely.

A serializer can be constructed with either an internal representation, a
transport representation or both. Internal as well as transport representation
can be queried after construction. To define the transition between
representations a specific serializer needs to implement the to_instance,
to_representation and merge methods. When both representations are given at
construction, the merge method is called.
'''


class Serializer:
    def __init__(self,
                 representation=None,
                 instance=None):

        has_representation = representation is not None
        has_instance = instance is not None

        if has_representation and has_instance:
            self._instance = self.merge(instance, representation)
            self._representation = None

        elif has_representation:
            def validate_if_has_validator(obj):
                if hasattr(self.__class__, 'validator'):
                    return self.__class__.validator.validate(obj)
                else:
                    return obj

            self._instance = None
            self._representation = validate_if_has_validator(representation)

        elif has_instance:
            self._instance = instance
            self._representation = None

        else:
            raise Exception('Need either instance or representation or both.')

    @property
    def instance(self):
        return (self._instance
                if self._instance is not None
                else self.to_instance(self._representation))

    @property
    def representation(self):
        return (self._representation
                if self._representation is not None
                else self.to_representation(self._instance))

    def to_instance(self, representation):
        raise Exception('implement me')

    def to_representation(self, instance):
        raise Exception('implement me')

    def merge(self, instance, representation):
        raise Exception('implement me')



from sys import stderr, stdout

from papyru.logger_types import Message, Sequence, Trace


def make_stdout_sink():
    return StdoutSink()


class StdoutSink:
    def _choose_stream(item):
        IS_ERROR_MAP = {
            Message.SUCCESS: False,
            Message.INFO: False,
            Message.WARNING: False,
            Message.FAILURE: True,
            Message.CRITICAL: True,
        }

        if isinstance(item, Message):
            return (stderr
                    if IS_ERROR_MAP[item.severity]
                    else stdout)
        elif isinstance(item, Trace):
            return stderr
        elif isinstance(item, Sequence):
            if item.resolution == Message.SUCCESS:
                return stdout
            else:
                return stderr
        else:
            raise NotImplementedError()

    def commit(self, item):
        stream = StdoutSink._choose_stream(item)
        stream.write(_encode_text(item))
        stream.write('\n')
        stream.flush()


_DECORATION_MAP = {
    Message.SUCCESS: '✓ ',
    Message.INFO: '- ',
    Message.WARNING: '⚠ ',
    Message.FAILURE: '✗ ',
    Message.CRITICAL: '✗✗✗ ',
}


def format_summary(item):
    lines = _encode_text(item).splitlines()

    if len(lines) == 0:
        return '(empty)'
    elif len(lines) == 1:
        return lines[0]
    else:
        return '%s [...] %s' % (lines[0], lines[-1])


def _encode_text(item, indent=0):
    indent_str = '  ' * indent

    if isinstance(item, Message):
        return '%s%s %s' % (indent_str,
                            _DECORATION_MAP[item.severity],
                            item.message)
    elif isinstance(item, Sequence):
        sub = '\n'.join(map(lambda i: _encode_text(i, indent + 1),
                            item.items))

        if len(sub) > 0:
            sub = '\n%s' % sub

        return '%s%s%s%s' % (indent_str,
                             _DECORATION_MAP[item.resolution],
                             item.description,
                             sub)
    elif isinstance(item, Trace):
        sub = '\n'.join(map(lambda f: _encode_text(f, indent),
                            item.frames))

        return '%s\n%s' % (sub, _encode_text(item.summary, indent))
    elif isinstance(item, Trace.Summary):
        return '%s✗ %s: %s' % (indent_str,
                               item.description,
                               item.exception)
    elif isinstance(item, Trace.Frame):
        return '%s↪ %s, %d: %s' % (indent_str,
                                   item.filename,
                                   item.linenumber,
                                   item.line)
    else:
        raise NotImplementedError()



from typing import Callable

from papyru.validation import CerberusValidator
from papyru.varspool.types import JobStatus


class VarspoolJobProcessorConfig:
    '''
    Configuration used by the base command scheduling varspool jobs.

    The configuration consists of the following keys:

    - queue_url: The URL to the varspool queue.
    - max_parallel_job_count: Maximum count of processes to start concurrently.
    - max_runtime_in_minutes: Duration in minutes for how long new jobs should
      be dispatched.
    - loop_cooldown_in_seconds (`Optional`): Indicates how long the job
      processor waits when there are no new jobs in the queue, when the queue
      is fully busy or when all jobs are dispatched.
      `Default`: 30.
    - failure_backoff_in_seconds (`Optional`): Indicates how many seconds the
      job processor should spend to wait when a failure occurred which makes
      the job processor unable to continue.
      `Default`: 600 (= 10 minutes).
    - job_handler: A function which actually does the hard work.
      It receives a `Job` object and should return a `JobResult` object.
    - abort_status: The status to which a job is resetted when the runtime
      is exhausted, or interrupted via SIGTERM.
      `Default`: open.
    '''

    DEFAULT_LOOP_COOLDOWN_IN_SECONDS = 30
    DEFAULT_FAILURE_BACKOFF_IN_SECONDS = 600
    DEFAULT_ABORT_STATUS = 'open'

    _CONFIG_SCHEMA = {
        'schema': {
            'queue_url': {
                'type': 'string',
                'required': True,
            },
            'max_parallel_job_count': {
                'type': 'integer',
                'required': True,
                'min': 1,
            },
            'max_runtime_in_minutes': {
                'type': 'integer',
                'required': True,
                'min': 0,
            },
            'loop_cooldown_in_seconds': {
                'type': 'integer',
                'min': 0,
            },
            'failure_backoff_in_seconds': {
                'type': 'integer',
                'min': 0,
            },
            'job_handler': {
                'required': True
            },
            'abort_status': {
                'type': 'string',
            },
        }
    }

    def __init__(self, config: dict):
        CerberusValidator(self._CONFIG_SCHEMA).validate(config)

        self._queue_url = config['queue_url']
        self._max_parallel_job_count = config['max_parallel_job_count']
        self._max_runtime_in_minutes = config['max_runtime_in_minutes']
        self._loop_cooldown_in_seconds = config.get(
            'loop_cooldown_in_seconds',
            VarspoolJobProcessorConfig.DEFAULT_LOOP_COOLDOWN_IN_SECONDS)
        self._failure_backoff_in_seconds = config.get(
            'failure_backoff_in_seconds',
            VarspoolJobProcessorConfig.DEFAULT_FAILURE_BACKOFF_IN_SECONDS)
        self._job_handler = config['job_handler']
        self._abort_status = JobStatus(
            config.get('abort_status',
                       VarspoolJobProcessorConfig.DEFAULT_ABORT_STATUS))

    @property
    def queue_url(self):
        return self._queue_url

    @property
    def max_parallel_job_count(self) -> int:
        return self._max_parallel_job_count

    @property
    def max_runtime_in_minutes(self) -> int:
        return self._max_runtime_in_minutes

    @property
    def loop_cooldown_in_seconds(self) -> int:
        return self._loop_cooldown_in_seconds

    @property
    def failure_backoff_in_seconds(self) -> int:
        return self._failure_backoff_in_seconds

    @property
    def job_handler(self) -> Callable:
        return self._job_handler

    @property
    def abort_status(self) -> JobStatus:
        return self._abort_status



from setuptools import setup

setup(
    name='papyru',
    version='1.9.8',
    description=(
        'minimal REST library with OpenAPI-based validation for django'),
    author='puzzle & play GmbH',
    author_email='papyru@puzzleandplay.de',
    url='https://www.fotopuzzle.de/',
    license='AGPLv3',
    platforms=['any'],
    packages=['papyru', 'papyru.static', 'papyru.varspool'],
    package_data={},
    install_requires=[
        'Cerberus',
        'Django',
        'jsonschema',
        'pyyaml',
        'requests',
        'lxml',
        'python-dateutil',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    scripts=[
        'bin/generate_jsonschema.py'
    ]
)



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from papyru.utils import PAPEnum


class JobStatus(PAPEnum):
    OPEN = 'open'
    IN_PROGRESS = 'in_progress'
    FAILED_TEMPORARILY = 'failed_temporarily'
    FAILED = 'failed'
    DONE = 'done'


@dataclass
class JobStatusEntry:
    status: JobStatus
    checksum: Optional[str] = None
    date: Optional[datetime] = None
    data: Optional[dict] = None

    @staticmethod
    def from_dict(entry_dict: dict):
        date = (datetime.fromisoformat(entry_dict['date'])
                if entry_dict.get('date', None) is not None
                else None)
        return JobStatusEntry(status=JobStatus(entry_dict['status']),
                              checksum=entry_dict.get('checksum'),
                              date=date,
                              data=entry_dict.get('data'))

    def to_dict(self):
        return {
            'status': str(self.status),
            **({'checksum': self.checksum}
               if self.checksum is not None else {}),
            **({'date': self.date.isoformat()}
               if self.date is not None else {}),
            **({'data': self.data} if self.data is not None else {})
        }


@dataclass
class JobStatusHistory:
    location: str
    items: List[JobStatusEntry]

    @staticmethod
    def from_dict(history_dict: dict):
        return JobStatusHistory(location=history_dict['location'],
                                items=list(map(JobStatusEntry.from_dict,
                                               history_dict['items'])))

    def to_dict(self):
        return {
            'location': self.location,
            'items': list(map(lambda it: it.to_dict(),
                              self.items)),
        }


@dataclass
class Job:
    id: int
    status_history: JobStatusHistory

    @staticmethod
    def from_dict(job_dict: dict):
        return Job(id=job_dict['id'],
                   status_history=JobStatusHistory.from_dict(
                       job_dict['status_history']))

    def to_dict(self):
        return {
            'id': self.id,
            'status_history': self.status_history.to_dict(),
        }


class JobResult:
    def __init__(self,
                 status: JobStatus,
                 data: dict):
        self.status = status
        self.data = data

    def __str__(self):
        return ('%s(status=%s, data=%s)'
                % (type(self).__name__, self.status, self.data))

    def to_status_entry(self) -> JobStatusEntry:
        return JobStatusEntry(status=self.status,
                              data=self.data)



from .auth import parse_authorization_headers, require_basic_auth
from .context import JSONContext, PlainContext, RequestContext, XMLContext
from .logger import (LogSequence, log_fail, log_info, log_item, log_ok,
                     log_trace, log_warn)
from .middleware import CommonMiddleware
from .problem import Problem
from .profiler import (DummyProfiler, Profiler, ReportCSVDetails, ReportList,
                       ReportPlotDetails, should_profile_request)
from .resource import Resource
from .serializer import Serializer
from .utils import limited_runtime
from .validation import CerberusValidator, JSONSchemaValidator, Validator
from .xml_helper import XMLHelper

__all__ = [
    'require_basic_auth', 'parse_authorization_headers',
    'PlainContext', 'RequestContext', 'JSONContext', 'XMLContext',
    'LogSequence', 'log_fail', 'log_info', 'log_item', 'log_trace', 'log_ok',
    'log_warn',
    'CommonMiddleware',
    'Problem',
    'Resource',
    'Serializer',
    'CerberusValidator', 'JSONSchemaValidator',
    'Validator',
    'XMLHelper',
    'DummyProfiler', 'Profiler', 'should_profile_request',
    'ReportList', 'ReportCSVDetails', 'ReportPlotDetails',
    'limited_runtime'
]



import mimetypes
import os
from collections import namedtuple
from enum import Enum
from wsgiref.util import FileWrapper

from django.conf import settings

from papyru import JSONContext, PlainContext, Problem, Resource

DEFAULT_ALLOW_LIST_STATIC_FILES = False


Kind = Enum('PapyruStaticKind', ['FILE', 'DIRECTORY'])

Candidate = namedtuple('PapyruStaticCandidate', ['path',
                                                 'kind',
                                                 'mimetype'])


class StaticFilesList(Resource):
    def get(self, request, file_name):
        candidate = _lookup_candidate(file_name)

        if candidate.kind == Kind.FILE:
            with open(candidate.path, 'rb') as static_file:
                return PlainContext(candidate.mimetype).response(
                    FileWrapper(static_file))
        else:
            if not _may_list_files():
                raise Problem.forbidden()
            listing = os.listdir(candidate.path)
            return JSONContext().response({'items': listing})

    def head(self, request, file_name):
        candidate = _lookup_candidate(file_name)

        if candidate.kind == Kind.FILE:
            return PlainContext(candidate.mimetype).response()
        else:
            if not _may_list_files():
                raise Problem.forbidden()

            return PlainContext('application/json').response()


def _may_list_files():
    return getattr(settings,
                   'PAPYRU_STATIC_FILES_ALLOW_LISTING',
                   DEFAULT_ALLOW_LIST_STATIC_FILES)


def _lookup_candidate(file_name):
    paths = map(lambda directory: os.path.join(directory, file_name),
                settings.PAPYRU_STATIC_FILES_DIRS)

    candidates = set(filter(lambda it: os.path.exists(it), paths))

    if len(candidates) < 1:
        raise Problem.not_found(file_name)

    if len(candidates) > 1:
        raise Problem.internal_error(
            'static file "%s" is ambigous: %s' % (file_name, candidates))

    path = next(iter(candidates))
    is_dir = os.path.isdir(path)

    return Candidate(path=path,
                     mimetype=(mimetypes.guess_type(path)[0]
                               if not is_dir
                               else None),
                     kind=(Kind.FILE if not is_dir else Kind.DIRECTORY))



import logging
import re
from contextlib import contextmanager
from datetime import datetime
from enum import Enum, unique
from os import getpid
from unittest.mock import patch
from uuid import uuid4

from papyru.logger import log_info

_logger = logging.getLogger(__name__)


@unique
class PAPEnum(Enum):
    def __str__(self):
        return self.value

    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


@contextmanager
def limited_runtime(timeout):
    '''
    Checks if runtime is left.

    Example:

        .. code-block:: python

            with limited_runtime(
                    datetime.timedelta(minutes=MAX_RUNTIME_MINUTES)
            ) as has_runtime_left:
                while has_runtime_left():
                    do_something()
                    sleep(1)
    '''

    start_time = datetime.now()

    def has_runtime_left():
        return (datetime.now() - start_time) < timeout

    try:
        yield has_runtime_left
    finally:
        pass


def parse_bool(value):
    if isinstance(value, bool):
        return value
    elif value is None:
        return False
    elif isinstance(value, str):
        return value.lower() in ('true', '1', 'yes')
    elif isinstance(value, int):
        return value == 1
    elif isinstance(value, float):
        return int(value) == 1
    else:
        raise TypeError('cannot parse bool from "%s".' % value)


def setup_request_ids(ignored_domains=[]):
    if setup_request_ids._already_patched:
        return
    else:
        setup_request_ids._already_patched = True

    import requests

    ignore_patterns = set(map(
        lambda d: re.compile('^http[s]?[:][/][/]%s.*$' % re.escape(d)),
        ignored_domains))

    class PatchedSession(requests.sessions.Session):
        def prepare_request(self, request):
            if any(map(lambda p: p.match(request.url),
                       ignore_patterns)):
                return super().prepare_request(request)

            request.headers = {**(request.headers or {}),
                               'pap-request-id': str(uuid4())}

            log_info('sending request %s: %s %s' % (
                request.headers.get('pap-request-id'),
                request.method,
                request.url))

            return super().prepare_request(request)

    requests.sessions.Session = PatchedSession


setup_request_ids._already_patched = False


def silent_log_commit(self, item):
    pass


def make_silent_testcase(TestCaseClass):
    class Impl(TestCaseClass):
        @patch('papyru.logger_stdout.StdoutSink.commit', silent_log_commit)
        def run(self, *args, **kwargs):
            logger = logging.getLogger()
            _loglevel = logger.getEffectiveLevel()

            try:
                logger.setLevel(logging.CRITICAL + 1)

                result = super().run(*args, **kwargs)

                logger.setLevel(_loglevel)

                return result

            finally:
                logger.setLevel(_loglevel)
    return Impl


def log(message, level='info'):
    getattr(_logger, level)('[%d]: %s' % (getpid(), message))
